"""
  Defines the errors that are returned from Guard
"""


class GuardError(Exception):
    """General exception to cover all Guard exceptions"""


class UnknownError(GuardError):
    """Raised when having an unknown error is encountered"""


class ParseError(GuardError, ValueError):
    """Raised when having an issue parsing rules"""


class MissingValueError(ParseError, NameError):
    """There was no variable or value object to resolve"""



"""
  CloudFormation Guard

  This library converts CloudFormation Guard into a Python
  library.  It leverages the Guard API of run_checks
  to validate a payload using a string version of the rules
"""

from .api import run_checks

from .interface import (
    FileReport,
    RuleReport,
    Messages,
    ClauseReport,
    UnaryReport,
    BinaryReport,
    UnaryComparison,
    UnaryCheck,
    GuardClauseReport,
    BinaryCheck,
    BinaryComparison,
)



"""
    Test cfn_guard_rs
"""
from unittest.mock import patch
import json
import yaml
import pytest
import cfn_guard_rs.errors
from cfn_guard_rs import (
    run_checks,
    FileReport,
    RuleReport,
    Messages,
    ClauseReport,
    UnaryReport,
    BinaryReport,
    UnaryComparison,
    UnaryCheck,
    GuardClauseReport,
    BinaryCheck,
    BinaryComparison,
)


@pytest.mark.parametrize(
    "template,rules,expected",
    [
        (
            "python/tests/fixtures/templates/s3_bucket_name_valid.yaml",
            "python/tests/fixtures/rules/s3_bucket_name.guard",
            FileReport(
                name="",
                metadata={},
                status="PASS",
                data_from="cfn_guard_rs",
                rules_from="cfn_guard_rs",
                not_compliant=[],
                not_applicable=[],
                compliant=["default"],
            ),
        ),
        (
            "python/tests/fixtures/templates/s3_bucket_name_invalid.yaml",
            "python/tests/fixtures/rules/s3_bucket_name.guard",
            FileReport(
                name="",
                metadata={},
                status="FAIL",
                data_from="cfn_guard_rs",
                rules_from="cfn_guard_rs",
                not_compliant=[
                    ClauseReport(
                        Rule=RuleReport(
                            name="default",
                            metadata={},
                            messages=Messages(),
                            checks=[
                                ClauseReport(
                                    Clause=GuardClauseReport(
                                        Unary=UnaryReport(
                                            context=" BucketName IS STRING  ",
                                            messages=Messages(
                                                custom_message="",
                                                error_message=(
                                                    "Check was not compliant as "
                                                    "property [/Resources/Bucket/Properties"
                                                    "/BucketName[L:0,C:0]] was not string."
                                                ),
                                            ),
                                            check=UnaryCheck(
                                                Resolved=UnaryComparison(
                                                    value={
                                                        "path": (
                                                            "/Resources/Bucket/"
                                                            "Properties/BucketName"
                                                        ),
                                                        "value": 1,
                                                    },
                                                    comparison=("IsString", False),
                                                )
                                            ),
                                        ),
                                    ),
                                ),
                            ],
                        ),
                    )
                ],
                not_applicable=[],
                compliant=[],
            ),
        ),
        (
            "python/tests/fixtures/templates/s3_bucket_public_access_valid.yaml",
            "python/tests/fixtures/rules/s3_bucket_public_access.guard",
            FileReport(
                name="",
                metadata={},
                status="PASS",
                data_from="cfn_guard_rs",
                rules_from="cfn_guard_rs",
                not_compliant=[],
                not_applicable=[],
                compliant=["S3_BUCKET_LEVEL_PUBLIC_ACCESS_PROHIBITED"],
            ),
        ),
        (
            "python/tests/fixtures/templates/s3_bucket_public_access_invalid.yaml",
            "python/tests/fixtures/rules/s3_bucket_public_access.guard",
            FileReport(
                name="",
                metadata={},
                status="FAIL",
                data_from="cfn_guard_rs",
                rules_from="cfn_guard_rs",
                not_compliant=[
                    ClauseReport(
                        Rule=RuleReport(
                            name="S3_BUCKET_LEVEL_PUBLIC_ACCESS_PROHIBITED",
                            metadata={},
                            messages=Messages(),
                            checks=[
                                ClauseReport(
                                    Clause=GuardClauseReport(
                                        Binary=BinaryReport(
                                            context=(
                                                " %s3_buckets_level_public_access_"
                                                "prohibited[*].Properties."
                                                "PublicAccessBlockConfiguration."
                                                'BlockPublicAcls EQUALS  "true"'
                                            ),
                                            messages=Messages(
                                                custom_message="",
                                                error_message=(
                                                    "Check was not compliant as property"
                                                    " value [Path=/Resources/Bucket/Properties/"
                                                    "PublicAccessBlockConfiguration/BlockPublicAcls"
                                                    '[L:0,C:0] Value="false"] not equal to value '
                                                    '[Path=[L:0,C:0] Value="true"].'
                                                ),
                                            ),
                                            check=BinaryCheck(
                                                Resolved=BinaryComparison(
                                                    from_={
                                                        "path": (
                                                            "/Resources/Bucket/Properties/"
                                                            "PublicAccessBlockConfiguration/"
                                                            "BlockPublicAcls"
                                                        ),
                                                        "value": "false",
                                                    },
                                                    to_={"path": "", "value": "true"},
                                                    comparison=["Eq", False],
                                                )
                                            ),
                                        ),
                                    ),
                                ),
                            ],
                        ),
                    )
                ],
                not_applicable=[],
                compliant=[],
            ),
        ),
    ],
)
def test_run_checks(template, rules, expected):
    """Test transactions against run_checks"""
    with open(template, encoding="utf8") as file:
        template_str = yaml.safe_load(file)
    with open(rules, encoding="utf8") as file:
        rules = file.read()

    result = run_checks(template_str, rules)
    print(result)
    print(expected)
    assert result == expected


@pytest.mark.parametrize(
    "template,rules,error,parent_error,message",
    [
        (
            "python/tests/fixtures/templates/s3_bucket_public_access_valid.yaml",
            "python/tests/fixtures/rules/invalid_missing_value.guard",
            cfn_guard_rs.errors.MissingValueError,
            NameError,
            (
                "Could not resolve variable by name "
                "ecs_task_definition_execution_role_arn across scopes"
            ),
        ),
        (
            "python/tests/fixtures/templates/s3_bucket_public_access_valid.yaml",
            "python/tests/fixtures/rules/invalid_format.guard",
            cfn_guard_rs.errors.ParseError,
            ValueError,
            (
                "Parser Error when parsing `Parsing Error Error parsing file  "
                "at line 1 at column 17, when handling , fragment "
                "{\n    Properties\n        BucketName is_string\n    }\n}`"
            ),
        ),
    ],
)
def test_run_checks_errors(template, rules, error, parent_error, message):
    """Test transactions against run_checks"""
    with open(template, encoding="utf8") as file:
        template_str = yaml.safe_load(file)
    with open(rules, encoding="utf8") as file:
        rules = file.read()

    with pytest.raises(error, match=message):
        run_checks(template_str, rules)

    with pytest.raises(parent_error, match=message):
        run_checks(template_str, rules)

    with pytest.raises(cfn_guard_rs.errors.GuardError, match=message):
        run_checks(template_str, rules)


@patch("cfn_guard_rs.api.run_checks_rs")
def test_json_decode_error(mock_run_check_rs):
    """Test JSON Decode errors"""

    mock_run_check_rs.return_value = '{ "bad": }'

    with pytest.raises(json.JSONDecodeError):
        run_checks("{}", "")


def test_run_unknown_errors():
    """Test unknown errors"""

    template_filename = "python/tests/fixtures/templates/s3_bucket_name_valid.yaml"
    rules_filename = "python/tests/fixtures/rules/s3_bucket_name.guard"
    with open(template_filename, encoding="utf8") as file:
        template_str = yaml.safe_load(file)
    with open(rules_filename, encoding="utf8") as file:
        rules = file.read()

    with patch("json.loads") as mock_method:
        mock_method.side_effect = Exception("test")
        with pytest.raises(cfn_guard_rs.errors.UnknownError):
            run_checks(template_str, rules)



"""
  Defines the non-verbose output format returned
  from run_checks
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, List, Dict, Sequence, Tuple

# pylint: disable=missing-class-docstring,missing-function-docstring,invalid-name

@dataclass(eq=True, frozen=True)
class Messages:
    """Messages Output"""

    custom_message: str | None = field(default=None)
    error_message: str | None = field(default=None)

    @classmethod
    def from_object(cls, obj) -> "Messages":
        return cls(
            custom_message=obj.get("custom_message"),
            error_message=obj.get("error_message"),
        )


@dataclass(eq=True, frozen=True)
class RuleReport:
    """Messages Output"""

    name: str = field()
    metadata: Dict[str, Any] = field()
    messages: Messages = field()
    checks: Sequence[ClauseReport] = field()

    @classmethod
    def from_object(cls, obj) -> "RuleReport" | None:
        if obj is None:
            return None

        return cls(
            name=obj.get("name"),
            metadata=obj.get("metadata"),
            messages=Messages.from_object(obj.get("messages")),
            checks=ClauseReport.from_array(obj.get("checks")),
        )


@dataclass(eq=True, frozen=True)
class GuardBlockReport:
    """Guard Block Report"""

    context: str = field()
    messages: Messages = field()
    unresolved: Any = field()

    @classmethod
    def from_object(cls, obj):
        if obj is None:
            return obj

        return cls(
            context=obj.get("context"),
            messages=Messages.from_object(obj.get("messages")),
            unresolved=obj.get("resolved"),
        )


@dataclass(eq=True, frozen=True)
class DisjunctionsReport:
    """Disjunctions"""

    checks: ClauseReport = field()

    @classmethod
    def from_object(cls, obj):
        if obj is None:
            return obj

        return cls(checks=ClauseReport.from_object(obj.get("checks")))

@dataclass(eq=True, frozen=True)
class UnaryComparison:
    """Unary Comparison"""

    value: Any = field()
    comparison: Tuple[str, bool] = field()

    @classmethod
    def from_object(cls, obj):
        return cls(
            value=obj.get("value"),
            comparison=tuple(obj.get("comparison")),
        )


@dataclass(eq=True, frozen=True)
class UnResolved:
    """Unresolved"""

    traversed_to: Any = field()
    remaining_query: Any = field()
    reason: Any = field()

    @classmethod
    def from_object(cls, obj):
        if obj is None:
            return None
        return cls(
            traversed_to=obj.get("traversed_to"),
            remaining_query=obj.get("remaining_query"),
            reason=obj.get("reason"),
        )


@dataclass(eq=True, frozen=True)
class ValueUnResolved:
    value: Any = field()
    comparison: Any = field()

    @classmethod
    def from_object(cls, obj):
        return cls(
            value=obj.get("value"),
            comparison=obj.get("comparison"),
        )


@dataclass(eq=True, frozen=True)
class UnaryCheck:
    """Unary Check"""

    Resolved: UnaryComparison | None = field(default=None)
    UnResolved: UnResolved | None = field(default=None)
    UnresolvedContext: Any | None = field(default=None)

    @classmethod
    def from_object(cls, obj):
        if obj is None:
            return obj

        return cls(
            Resolved=UnaryComparison.from_object(obj.get("Resolved")),
            UnResolved=UnResolved.from_object(obj.get("UnResolved")),
            UnresolvedContext=obj.get("UnresolvedContext"),
        )


@dataclass(eq=True, frozen=True)
class UnaryReport:
    """Unary Report"""

    context: str = field()
    messages: Messages = field()
    check: UnaryCheck = field()

    @classmethod
    def from_object(cls, obj):
        if obj is None:
            return None

        return cls(
            context=obj.get("context"),
            messages=Messages.from_object(obj.get("messages")),
            check=UnaryCheck.from_object(obj.get("check")),
        )


@dataclass(eq=True, frozen=True)
class BinaryComparison:
    from_: Any = field()
    to_: Any = field()
    comparison: Any = field()

    @classmethod
    def from_object(cls, obj):
        if obj is None:
            return None
        return cls(
            from_=obj.get("from"), to_=obj.get("to"), comparison=obj.get("comparison")
        )


@dataclass(eq=True, frozen=True)
class InComparison:
    def __init__(self, **kwargs) -> None:
        self.from_ = kwargs["from"]
        self.to_ = kwargs["to"]
        self.comparison = kwargs["comparison"]

    @classmethod
    def from_object(cls, obj) -> "InComparison" | None:
        return cls(
            from_=obj.get("from"),
            to_=obj.get("to"),
            comparison=obj.get("comparison"),
        )


@dataclass(eq=True, frozen=True)
class BinaryCheck:
    Resolved: BinaryComparison | None = field(default=None)
    UnResolved: UnResolved | None = field(default=None)
    InResolved: Any = field(default=None)

    @classmethod
    def from_object(cls, obj):
        return cls(
            Resolved=BinaryComparison.from_object(obj.get("Resolved")),
            UnResolved=UnResolved.from_object(obj.get("UnResolved")),
            InResolved=obj.get("InResolved"),
        )


@dataclass(eq=True, frozen=True)
class BinaryReport:
    context: str = field()
    messages: Messages = field()
    check: BinaryCheck = field()

    @classmethod
    def from_object(cls, obj):
        if obj is None:
            return obj
        return cls(
            context=obj.get("context"),
            messages=Messages.from_object(obj.get("messages")),
            check=BinaryCheck.from_object(obj.get("check")),
        )


@dataclass(eq=True, frozen=True)
class GuardClauseReport:
    """Guard Clause Report"""

    Unary: UnaryReport | None = field(default=None)
    Binary: BinaryReport | None = field(default=None)

    @classmethod
    def from_object(cls, obj):
        if obj is None:
            return obj

        return cls(
            Unary=UnaryReport.from_object(obj.get("Unary")),
            Binary=BinaryReport.from_object(obj.get("Binary")),
        )


@dataclass(eq=True, frozen=True)
class ClauseReport:
    """Clause Report"""

    Rule: RuleReport | None = field(default=None)
    Disjunctions: DisjunctionsReport | None = field(default=None)
    GuardBlock: GuardBlockReport | None = field(default=None)
    Clause: GuardClauseReport | None = field(default=None)

    @classmethod
    def from_object(cls, obj):
        return cls(
            Rule=RuleReport.from_object(obj.get("Rule")),
            Disjunctions=DisjunctionsReport.from_object(obj.get("Disjunctions")),
            GuardBlock=GuardBlockReport.from_object(obj.get("GuardBlock")),
            Clause=GuardClauseReport.from_object(obj.get("Clause")),
        )

    @classmethod
    def from_array(cls, items):
        results = []
        for item in items:
            results.append(ClauseReport.from_object(item))

        return results


# pylint: disable=too-many-instance-attributes
@dataclass(eq=True, frozen=True)
class FileReport:
    name: str = field()
    metadata: Dict[str, Any] = field()
    status: str = field()
    not_compliant: List[ClauseReport] = field()
    not_applicable: List[str] = field()
    compliant: List[str] = field()
    data_from: str = field(default="cfn_guard_rs")
    rules_from: str = field(default="cfn_guard_rs")

    @classmethod
    def from_object(cls, obj) -> "FileReport":
        return cls(
            name=obj.get("name"),
            metadata=obj.get("metadata"),
            status=obj.get("status"),
            not_compliant=ClauseReport.from_array(obj.get("not_compliant")),
            not_applicable=obj.get("not_applicable"),
            compliant=obj.get("compliant"),
            data_from=obj.get("data_from", "cfn_guard_rs"),
            rules_from=obj.get("rules_from", "cfn_guard_rs"),
        )



"""
  Defines the api for cfn_guard_rs

  Supports running the non-verbose version of
  run_checks from CloudFormation Guard
"""
import logging
import json
from cfn_guard_rs.interface import FileReport

# pylint: disable=no-name-in-module
from cfn_guard_rs.cfn_guard_rs import (
    CfnGuardParseError,
    CfnGuardMissingValue,
    run_checks_rs,
)

from . import errors

LOG = logging.getLogger(__name__)


def run_checks(data: dict, rules: str) -> FileReport:
    """
    Executes run_checks against Guard

    Run non-verbose checks against data for rules

    Parameters
    ----------
    data : dict
        Data is the object being checked by guard
    rules : str
        A string representation of the ruels being used

    Returns
    -------
    DataOuptut
        DataOutput representation of the result of running guard
    """
    try:
        raw_output = run_checks_rs(json.dumps(data), rules, False)
        output = json.loads(raw_output)

        return FileReport.from_object(output)
    except json.JSONDecodeError as err:
        LOG.debug(
            "JSON decoding error when processing return value [%s] got error: %s",
            raw_output,
            err,
        )
        raise err
    except CfnGuardMissingValue as err:
        raise errors.MissingValueError(str(err))
    except CfnGuardParseError as err:
        raise errors.ParseError(str(err))
    except Exception as err:
        LOG.debug(
            "Received unknown exception [%s] while running checks, got error: %s",
            type(err),
            err,
        )
        raise errors.UnknownError(str(err))
